use crate::error::{ExecutionError, GeorgeError, GeorgeErrorKind, MemoryError};
use crate::instructions::{get_instruction, Instruction};
use crate::memory::{self, Mem};
use crate::types::{Byte, Word};
use std::process::exit;
use std::time::Duration;
use std::{
    str::FromStr,
    sync::{Arc, RwLock},
    thread::sleep,
};

#[derive(Clone, Copy)]
pub enum StatusFlag {
    Negative = 0b1000_0000,
    Overflow = 0b0100_0000,
    Brk = 0b0011_0000,
    BrkIrq = 0b0010_0000,
    Decimal = 0b0000_1000,
    IrqDisable = 0b0000_0100,
    Zero = 0b000_0010,
    Carry = 0b0000_0001,
}

pub struct Cpu {
    pub a: Byte,  // Accumulator Register
    pub x: Byte,  // X Register
    pub y: Byte,  // Y Register
    pub pc: Word, // Program Counter
    pub s: Byte,  // Stack Pointer
    pub p: Byte,  // Status Register
    pub irq: bool,
    pub nmi: bool,
    pub memory: Arc<RwLock<Mem>>,
    pub pending_cycles: usize,
    cycle_count: usize,
}

impl Cpu {
    pub fn new(memory: Arc<RwLock<Mem>>) -> Self {
        Cpu {
            a: 0x00,
            x: 0x00,
            y: 0x00,
            pc: 0x0000,
            s: 0xFF,
            p: 0b0010_0100,
            irq: false,
            nmi: false,
            memory,
            pending_cycles: 0,
            cycle_count: 0,
        }
    }
    pub fn reset(&mut self) -> Result<(), ExecutionError> {
        let reset_vector_pointer = self.read_word(0xFFFC)?;
        self.pc = reset_vector_pointer;
        self.pending_cycles = 8;
        Ok(())
    }
    pub fn read(&self, address: Word) -> Result<Byte, MemoryError> {
        let memory = match self.memory.try_read() {
            Ok(read) => read,
            Err(_) => {
                println!("Couldn't acquire read lock on memory in cpu thread");
                return Err(MemoryError::NoDataAtAddress);
            } // TODO: upgrade this error type to a `GeorgeError` and make an errorkind for thread lock errors
        };
        memory.read(address)
    }
    pub fn read_word(&self, address: Word) -> Result<Word, MemoryError> {
        let low_byte = self.read(address)?;
        let high_byte = self.read(address + 0x1)?;
        Ok((high_byte as u16) << 8 | (low_byte as u16))
    }

    fn stack_addr(&self) -> Word {
        // Dunno if this is necessary, i just don't like adding the 0x0100 every time
        0x0100 + self.s as u16
    }

    pub fn push_stack(&mut self, data: Byte) -> Result<(), ExecutionError> {
        self.s -= 0x1;
        self.write(self.stack_addr(), data)?;
        Ok(())
    }

    pub fn push_stack_word(&mut self, address: Word) -> Result<(), ExecutionError> {
        self.s -= 0x1;
        self.write(self.stack_addr(), address.to_le_bytes()[1])?; // Upper byte first
        self.s -= 0x1;
        self.write(self.stack_addr(), address.to_le_bytes()[0])?; // Lower byte second
        Ok(())
    }

    pub fn pop_stack(&mut self) -> Result<Byte, ExecutionError> {
        let byte = self.read(self.stack_addr())?;
        self.s += 0x1;
        Ok(byte)
    }

    pub fn pop_stack_word(&mut self) -> Result<Word, ExecutionError> {
        let low_byte = self.pop_stack()?;
        let high_byte = self.pop_stack()?;
        let word = ((high_byte as Word) << 8) + low_byte as u16;
        Ok(word)
    }

    pub fn set_flag(&mut self, flag: StatusFlag, value: bool) {
        self.p &= !(flag as Byte);
        if value {
            self.p |= flag as Byte
        }
    }

    pub fn get_flag(&self, flag: StatusFlag) -> bool {
        (self.p & flag as Byte) > 0
    }
    pub fn is_negative(&self, value: Byte) -> bool {
        value & 0b1000_0000 == 0b1000_0000
    }

    pub fn write(&mut self, address: Word, data: Byte) -> Result<(), MemoryError> {
        let mut memory = match self.memory.write() {
            Ok(write) => write,
            Err(_) => {
                println!("Couldn't acquire write lock on memory in cpu thread");
                return Err(MemoryError::NoDataAtAddress);
            } // TODO: upgrade this error type to a `GeorgeError` and make an errorkind for thread lock errors
        };
        memory.write(address, data)?;
        Ok(())
    }

    pub fn execute(&mut self) {
        self.cycle().unwrap();
        while self.pending_cycles != 0 {
            self.cycle().unwrap();
        }
    }

    pub fn wait_for_interrupt(&self) {
        unimplemented!()
    }

    // fn handle_interrupt(&mut self) -> Result<(), ExecutionError> {
    //     match self.get_flag(StatusFlag::IrqDisable) {
    //         // Check that interrupts aren't disabled
    //         true => Ok(self.execute()),
    //         false => {
    //             if self.irq {
    //                 let irq_vector = 0xFFFE;
    //                 match self.read_word(irq_vector) {
    //                     Err(error) => Err(ExecutionError::MemoryError(error)),
    //                     Ok(address) => {
    //                         let isr_address = address;
    //                         match self.read(isr_address) {
    //                             Ok(_opcode) => Ok(self.execute()),
    //                             Err(error) => Err(ExecutionError::MemoryError(error)),
    //                         }
    //                     }
    //                 }
    //             } else {
    //                 Ok(())
    //             }
    //         }
    //     }
    // }
    pub fn cycle(&mut self) -> Result<(), GeorgeError> {
        sleep(Duration::from_nanos(100));
        if self.pending_cycles == 0 {
            if self.nmi || (self.irq && !self.get_flag(StatusFlag::IrqDisable)) {
                let _ = self.push_stack_word(self.pc);
                self.set_flag(StatusFlag::BrkIrq, true);
                let _ = self.push_stack(self.p);
                self.set_flag(StatusFlag::IrqDisable, true);
                if self.nmi {
                    match self.read_word(0xFFFA) {
                        Ok(word) => self.pc = word,
                        Err(error) => {
                            return Err(GeorgeError {
                                kind: GeorgeErrorKind::Memory(error),
                                desc: String::from_str("Couldn't read NMI vector").unwrap(),
                            })
                        }
                    };
                    self.pending_cycles = 8;
                } else {
                    match self.read_word(0xFFFE) {
                        Ok(word) => self.pc = word,
                        Err(error) => {
                            return Err(GeorgeError {
                                kind: GeorgeErrorKind::Memory(error),
                                desc: String::from_str("Couldn't read IRQ vector").unwrap(),
                            })
                        }
                    };
                    self.pending_cycles = 7
                }
                self.nmi = false;
                self.irq = false;
            }
        } else {
            let opcode = match self.read(self.pc) {
                Ok(byte) => byte,
                Err(error) => {
                    return Err(GeorgeError {
                        desc: String::from_str("Failed to read the program counter!").unwrap(),
                        kind: GeorgeErrorKind::Memory(error),
                    })
                }
            };
            let instruction = get_instruction(opcode);
            match instruction {
                Instruction::Valid(valid_instruction) => {
                    println!("a: {a:#04x}, x: {x:#04x}, y: {y:#04x}, pc: {pc:#06x}, sp: {s:#04x}, sr: {p:#010b}, irq: {irq:?}, nmi: {nmi:?}", a = self.a, x = self.x, y = self.y, pc = self.pc, s = self.s, p = self.p, irq = self.irq, nmi = self.nmi);
                    println!(
                        "Instruction: {:?}, {:#04x}",
                        valid_instruction.opcode, opcode
                    );
                    println!("");
                    self.pc += 1;
                    match valid_instruction.opcode.call(self) {
                        Ok(_) => {
                            self.pending_cycles += valid_instruction.cycles as usize;
                        },
                        Err(error) => return Err(GeorgeError{
                            desc: format!("An IncompatibleAddrMode was used at address {:#06x} for instruction {:?}", 
                                          self.pc, valid_instruction).to_string(), kind: error})
                    };
                }
                Instruction::Invalid(invalid_instruction) => match invalid_instruction.opcode {
                    0x02 => {
                        let memory = match self.memory.try_read() {
                            Ok(read) => read,
                            Err(_) => {
                                println!("Couldn't acquire read lock on memory in cpu thread");
                                return Err(GeorgeError {
                                    kind: GeorgeErrorKind::Memory(MemoryError::Unwritable),
                                    desc: "Couldn't acquire read lock on memory in cpu thread"
                                        .to_string(),
                                });
                            }
                        };
                        memory.dump().unwrap();
                        exit(1);
                    }
                    _ => {
                        return Err(GeorgeError {
                            kind: GeorgeErrorKind::Execution(ExecutionError::InvalidInstruction),
                            desc: format!(
                        "An invalid instruction with opcode {:#04x} was called at address {:#06x}",
                        invalid_instruction.opcode, self.pc
                    ),
                        });
                    }
                },
            }
        }
        self.pending_cycles -= 1;
        self.cycle_count += 1;
        Ok(())
    }

    pub fn stop(&mut self) {
        unimplemented!()
    }
}
