use crate::error::{GeorgeError, GeorgeErrorKind, MappingError, MemoryError};
use crate::types::{Byte, Word};
use std::io::{self, Write};
use std::path::PathBuf;
use std::{fs::File, io::Read};

#[derive(Debug, Clone)]
pub struct MemMappedDevice {
    start: Word,
    end: Word,
    pages: usize,
    page: usize,
    data: Vec<Byte>,
    label: String,
}

impl MemMappedDevice {
    pub fn new(start: Word, end: Word, pages: usize, label: String) -> Self {
        Self {
            start,
            end,
            pages,
            page: 0,
            data: vec![0x00; (end as usize + 1 - start as usize) * pages],
            label,
        }
    }
    fn contains(&self, address: Word) -> bool {
        self.start <= address && self.end >= address
    }
    fn swap_page(&mut self, page: usize) -> Result<(), GeorgeError> {
        match page > self.pages {
            true => Err(GeorgeError {
                kind: GeorgeErrorKind::Memory(MemoryError::Unmapped), //TODO: should be MappingError::InvalidPage,
                desc: format!(
                    "{:?} tried to swap to a page outside of its range",
                    self.label
                ),
            }),
            false => {
                self.page = page;
                Ok(())
            }
        }
    }
    fn size(&self) -> Word {
        self.end - self.start + 1
    }
    fn translate_address(&self, address: Word) -> Word {
        (address - self.start) + self.size() * (self.page as Word)
    } // This needs to translate memory address from CPU land to local land, so
      // for rom an address like 0xFFFF needs to be translated to Page X, 0xFFF
}

#[derive(Debug, Clone)]
pub struct Mem {
    areas: Vec<MemMappedDevice>,
}

impl Mem {
    pub fn new(area: MemMappedDevice) -> Self {
        Self { areas: vec![area] }
    }
    pub fn dump(&self) -> io::Result<()> {
        let mut outfile = File::create("./coredump.bin")?;
        let mut data = Vec::new();
        for area in &self.areas {
            for byte in &area.data {
                data.push(byte.to_owned());
            }
        }
        outfile.set_len(0xFFFF)?;
        outfile.write_all(&data)?;
        Ok(())
    }
    pub fn add_area(&mut self, area: MemMappedDevice) -> Result<(), GeorgeError> {
        for existing_area in &self.areas {
            if existing_area.contains(area.end) || existing_area.contains(area.start) {
                return Err(GeorgeError {
                    desc: format!(
                        "Tried to assign an area already allocated to {:?}",
                        existing_area
                    ),
                    kind: GeorgeErrorKind::Mapping(MappingError::RegionOccupied),
                });
            }
        }
        self.areas.push(area);
        Ok(())
    }
    pub fn read(&self, address: Word) -> Result<Byte, MemoryError> {
        for area in &self.areas {
            if area.contains(address) {
                let translated_address = area.translate_address(address);
                match area.data.get(translated_address as usize) {
                    Some(data) => return Ok(*data),
                    None => return Err(MemoryError::NoDataAtAddress),
                }
            }
        }
        Err(MemoryError::Unmapped)
    }

    pub fn write(&mut self, address: Word, data: Byte) -> Result<(), MemoryError> {
        for area in self.areas.iter_mut() {
            if area.contains(address) {
                // println!("Writing to area {label}", label = area.label);
                let translated_address = area.translate_address(address);
                area.data[translated_address as usize] = data;
                return Ok(());
            };
        }
        Err(MemoryError::Unmapped)
    }

    pub fn read_from_bin(&mut self, f: File) -> Result<(), MemoryError> {
        let bytes = f.bytes();
        for (address, byte) in bytes.enumerate() {
            match byte {
                Ok(value) => self.write(address as Word, value)?,
                Err(_) => return Err(MemoryError::Unwritable),
            }
        }
        Ok(())
    }
}
